@@ -24,12 +24,9 @@ G_BEGIN_DECLS
 struct _GIOPConnection {
 	LINCConnection  parent;
 
-	GMutex         *incoming_mutex;
 	GIOPRecvBuffer *incoming_msg;
 	GList          *incoming_frags;
 
-	GMutex         *outgoing_mutex;
-
 	GIOPVersion     giop_version;
 
 	gpointer        orb_data;
