@@ -31,6 +31,7 @@ test_safe_socket_dir (const char *dirname)
 		return FALSE;
 	}
 	
+#ifndef __CYGWIN__
 	if (statbuf.st_uid != getuid ()) {
 		S_PRINT (("Owner of %s is not the current user\n", dirname));
 		return FALSE;
@@ -41,6 +42,7 @@ test_safe_socket_dir (const char *dirname)
 		S_PRINT (("Wrong permissions for %s\n", dirname));
 		return FALSE;
 	}
+#endif
 
 	return TRUE;
 }
@@ -62,6 +64,8 @@ scan_socket_dir (const char *dir, const char *prefix)
 	g_return_val_if_fail (prefix != NULL, NULL);
 	
 	dirh = opendir (dir);
+	if (!dirh)
+		return NULL;
 	prefix_len = strlen (prefix);
 
 	while ((dent = readdir (dirh))) {
@@ -91,21 +95,21 @@ scan_socket_dir (const char *dir, const char *prefix)
 	return cur_dir;
 }
 
-#define PATH_ROOT "/tmp"
-
 static void
 giop_tmpdir_init (void)
 {
+	const char *tmp_root;
 	char *dirname;
 	char *safe_dir = NULL;
 	long iteration = 0;
 
+	tmp_root = g_get_tmp_dir ();
 	dirname = g_strdup_printf ("orbit-%s",
 				   g_get_user_name ());
 	while (!safe_dir) {
 		char *newname;
 
-		safe_dir = scan_socket_dir (PATH_ROOT, dirname);
+		safe_dir = scan_socket_dir (tmp_root, dirname);
 		if (safe_dir) {
 			dprintf (GIOP, "Have safe dir '%s'", safe_dir);
 			linc_set_tmpdir (safe_dir);
@@ -113,7 +117,7 @@ giop_tmpdir_init (void)
 		}
 
 		if (iteration == 0)
-			newname = g_strconcat (PATH_ROOT, "/", dirname, NULL);
+			newname = g_strconcat (tmp_root, "/", dirname, NULL);
 		else {
 			struct {
 				guint32 a;
@@ -124,7 +128,7 @@ giop_tmpdir_init (void)
 					     ORBIT_GENUID_OBJECT_ID);
 
 			newname = g_strdup_printf (
-				"%s/%s-%4x", PATH_ROOT, dirname, id.b);
+				"%s/%s-%4x", tmp_root, dirname, id.b);
 		}
 
 		if (mkdir (newname, 0700) < 0) {
@@ -164,7 +168,7 @@ giop_tmpdir_init (void)
 		g_free (newname);
 
 		if (iteration == 1000)
-			g_error ("Cannot find a safe socket path in '%s'", PATH_ROOT);
+			g_error ("Cannot find a safe socket path in '%s'", tmp_root);
 	}
 
 	g_free (safe_dir);
@@ -285,7 +289,7 @@ static int      corba_wakeup_fds[2];
 #define WAKEUP_POLL  corba_wakeup_fds [0]
 #define WAKEUP_WRITE corba_wakeup_fds [1]
 static GSource *giop_main_source = NULL;
-static GThread *giop_incoming_thread = NULL;
+static GIOPThread *giop_main_thread = NULL;
 
 static gboolean
 giop_mainloop_handle_input (GIOChannel     *source,
@@ -310,13 +314,12 @@ giop_init (gboolean threaded, gboolean blank_wire_data)
 
 	if (threaded) {
 		GIOPThread *tdata;
-		GError     *error = NULL;
 
 		/* FIXME: should really cleanup with descructor */
 		giop_tdata_private = g_private_new (NULL);
 
-		tdata = giop_thread_new ( /* main thread */
-			g_main_context_default ());
+		giop_main_thread = tdata = giop_thread_new (
+			g_main_context_default ()); /* main thread */
 
 		if (pipe (corba_wakeup_fds) < 0) /* cf. g_main_context_init_pipe */
 			g_error ("Can't create CORBA main-thread wakeup pipe");
@@ -327,16 +330,6 @@ giop_init (gboolean threaded, gboolean blank_wire_data)
 			giop_mainloop_handle_input, NULL);
 		
 		g_private_set (giop_tdata_private, tdata);
-
-		giop_incoming_thread =
-			g_thread_create_full (
-				giop_recv_thread_fn,
-				NULL, /* data */
-				0, TRUE, FALSE, 
-				G_THREAD_PRIORITY_NORMAL,
-				&error);
-		if (!giop_incoming_thread || error)
-			g_error ("Failed to create ORB worker thread");
 	}
 
 	giop_tmpdir_init ();
@@ -353,9 +346,10 @@ wakeup_mainloop (void)
 	char c = 'A'; /* magic */
 	int  res;
 	while ((res = write (WAKEUP_WRITE, &c, sizeof (c))) < 0  &&
-	       errno == EAGAIN);
+	       (errno == EAGAIN || errno == EINTR));
 	if (res < 0)
-		g_error ("Failed to write to wakeup socket");
+		g_error ("Failed to write to GIOP wakeup socket %d 0x%x(%d) (%d)",
+			 res, errno, errno, WAKEUP_WRITE);
 }
 
 void
@@ -372,8 +366,8 @@ giop_thread_push_recv (GIOPMessageQueueEntry *ent)
 	g_mutex_lock (tdata->lock);
 
 	if (ent->async_cb) /* we need the recv buffer later */
-		tdata->recv_buffers = g_slist_append (
-			tdata->recv_buffers, ent->buffer);
+		tdata->reply_list = g_slist_append (
+			tdata->reply_list, ent->buffer);
 	/* else - someone already waiting on the stack */
 
 	/* wakeup thread ... */
@@ -387,28 +381,49 @@ giop_thread_push_recv (GIOPMessageQueueEntry *ent)
 	g_mutex_unlock (tdata->lock);
 }
 
+static GMainLoop *giop_main_loop = NULL;
+
+void
+giop_main_run (void)
+{
+	if (giop_is_threaded) {
+		g_assert (giop_thread_self () == giop_main_thread);
+		g_assert (giop_main_loop == NULL);
+		giop_main_loop = g_main_loop_new (NULL, TRUE);
+		g_main_loop_run (giop_main_loop);
+		g_main_loop_unref (giop_main_loop);
+		giop_main_loop = NULL;
+	} else
+		linc_main_loop_run ();
+}
+
 void
 giop_shutdown (void)
 {
 	if (!giop_is_threaded)
 		giop_connections_shutdown ();
-	g_main_loop_quit (linc_loop); /* break into the linc loop */
+
+	if (linc_loop) /* break into the linc loop */
+		g_main_loop_quit (linc_loop);
+	if (giop_main_loop)
+		g_main_loop_quit (giop_main_loop);
 
 	if (giop_is_threaded) {
 		if (!giop_thread_self ()->wake_context)
 			g_error ("Must shutdown ORB from main thread");
 
-		g_thread_join (giop_incoming_thread);
-		giop_incoming_thread = NULL;
+		if (giop_main_source) {
+			g_source_destroy (giop_main_source);
+			g_source_unref (giop_main_source);
+			giop_main_source = NULL;
+		}
 
-		g_source_destroy (giop_main_source);
-		g_source_unref (giop_main_source);
-		giop_main_source = NULL;
-
-		close (WAKEUP_WRITE);
-		close (WAKEUP_POLL);
-		WAKEUP_WRITE = -1;
-		WAKEUP_POLL = -1;
+		if (WAKEUP_WRITE >= 0) {
+			close (WAKEUP_WRITE);
+			close (WAKEUP_POLL);
+			WAKEUP_WRITE = -1;
+			WAKEUP_POLL = -1;
+		}
 	}
 }
 
@@ -417,55 +432,66 @@ typedef struct {
 	gpointer recv_buffer;
 } GIOPQueueEntry;
 
+void
+giop_thread_request_process (GIOPThread *tdata)
+{
+	GIOPQueueEntry *qe = NULL;
 
-struct _GIOPQueue {
-	GList           *entries;
-	GIOPQueueHandler handler_fn;
-	void           (*push_fn) (GIOPQueue      *queue,
-				   GIOPQueueEntry *qe);
-};
+	LINC_MUTEX_LOCK (tdata->lock);
+
+	if (tdata->request_queue) {
+		qe = tdata->request_queue->data;
+		tdata->request_queue = g_slist_delete_link
+			(tdata->request_queue, tdata->request_queue);
+	}
+	
+	LINC_MUTEX_UNLOCK (tdata->lock);
+
+	if (qe)
+		tdata->request_handler (qe->poa_object, qe->recv_buffer, NULL);
+}
 
 void
-giop_queue_push (GIOPQueue *queue,
-		 gpointer   poa_object,
-		 gpointer   recv_buffer)
+giop_thread_request_push (GIOPThread *tdata,
+			  gpointer   *poa_object,
+			  gpointer   *recv_buffer)
 {
 	GIOPQueueEntry *qe;
 
-	g_return_if_fail (queue != NULL);
+	g_return_if_fail (tdata != NULL);
+	g_return_if_fail (poa_object != NULL);
+	g_return_if_fail (recv_buffer != NULL);
 
 	qe = g_new (GIOPQueueEntry, 1);
 
-	qe->poa_object  = poa_object;
-	qe->recv_buffer = recv_buffer;
+	qe->poa_object  = *poa_object;
+	qe->recv_buffer = *recv_buffer;
+	*poa_object = NULL;
+	*recv_buffer = NULL;
 
-	/* FIXME: no locking */
-	queue->entries = g_list_append (queue->entries, qe);
+	LINC_MUTEX_LOCK (tdata->lock);
 
-	/* FIXME: no special casing */
-	if (queue->push_fn)
-		queue->push_fn (queue, qe);
+	tdata->request_queue = g_slist_append (tdata->request_queue, qe);
+
+	g_cond_signal (tdata->incoming);
+	if (tdata->wake_context)
+		wakeup_mainloop ();
+
+	LINC_MUTEX_UNLOCK (tdata->lock);
 }
 
-GIOPQueue *
-giop_queue_new (gpointer queue_handler)
+GIOPThread *
+giop_thread_get_main (void)
 {
-	GIOPQueue *queue = g_new0 (GIOPQueue, 1);
-
-	queue->handler_fn = queue_handler;
-	
-	return queue;
-}
-
-GIOPQueue *
-giop_queue_get_main (void)
-{
-	return NULL;
+	return giop_main_thread;
 }
 
 void
-giop_queue_free (GIOPQueue *queue)
+giop_thread_set_main_handler (gpointer request_handler)
 {
-	g_assert (queue->entries == NULL);
-	g_free (queue);
+	if (!giop_is_threaded)
+		return;
+	g_assert (giop_main_thread != NULL);
+
+	giop_main_thread->request_handler = request_handler;
 }
