@@ -220,7 +220,7 @@ ORBit_small_marshal_context (GIOPSendBuffer *send_buffer,
 	   efficiency of the current impl */
 	ORBit_ContextMarshalItem *mlist;
 
-	mlist = alloca (sizeof (ORBit_ContextMarshalItem) *
+	mlist = g_alloca (sizeof (ORBit_ContextMarshalItem) *
 			m_data->contexts._length);
 
 	tprintf (" context { ");
@@ -284,7 +284,8 @@ typedef struct {
 #define do_marshal_value(a,b,c)     \
 	ORBit_marshal_value   ((a),(gconstpointer *)(b),(c))
 #define do_demarshal_value(a,b,c,e) \
-	ORBit_demarshal_value ((c),(b),(a),(e))
+	if (ORBit_demarshal_value ((c),(b),(a),(e))) \
+		goto demarshal_exception
 
 static gboolean
 orbit_small_marshal (CORBA_Object           obj,
@@ -305,7 +306,7 @@ orbit_small_marshal (CORBA_Object           obj,
 	{
 		int          align;
 		int          len = sizeof (CORBA_unsigned_long) + m_data->name_len + 1;
-		guchar      *header = alloca (len + sizeof (CORBA_unsigned_long));
+		guchar      *header = g_alloca (len + sizeof (CORBA_unsigned_long));
 
 		*(CORBA_unsigned_long *) header = m_data->name_len + 1;
 		memcpy (header + sizeof (CORBA_unsigned_long),
@@ -321,7 +322,8 @@ orbit_small_marshal (CORBA_Object           obj,
 	}
 
 	send_buffer = giop_send_buffer_use_request (
-		cnx->giop_version, request_id, CORBA_TRUE,
+		cnx->giop_version, request_id, 
+		(m_data->flags & ORBit_I_METHOD_1_WAY) == 0,
 		obj->object_key, &op_vec, NULL);
 
 	if (!send_buffer)
@@ -400,7 +402,7 @@ orbit_small_demarshal (CORBA_Object           obj,
 	GIOPRecvBuffer *recv_buffer;
 	CORBA_ORB       orb = obj->orb;
 
-	recv_buffer = giop_recv_buffer_get (mqe, TRUE);
+	recv_buffer = giop_recv_buffer_get (mqe);
 	if (!recv_buffer) {
 		dprintf (MESSAGES, "No recv buffer ...\n");
 		return MARSHAL_SYS_EXCEPTION_INCOMPLETE;
@@ -470,6 +472,8 @@ orbit_small_demarshal (CORBA_Object           obj,
 			dump_arg (a, tc);
 
 			if (a->flags & ORBit_I_ARG_OUT)
+				/* this may read (&discard) uninitialized memory,
+				 * see 'foo' below. This is for simplicity. */
 				arg = *(gpointer *)args [i];
 			else
 				arg = args [i];
@@ -504,7 +508,7 @@ orbit_small_demarshal (CORBA_Object           obj,
 				} else if (a->flags & ORBit_I_ARG_INOUT) {
 					ORBit_freekids_via_TypeCode (tc, arg);
 					do_demarshal_value (recv_buffer, &arg, tc, orb);
-				} else
+				} else /* 'foo' - don't use the bogus 'arg' contents */
 					*(gpointer *)args [i] = p = ORBit_demarshal_arg (
 						recv_buffer, tc, obj->orb);
 
@@ -526,6 +530,12 @@ orbit_small_demarshal (CORBA_Object           obj,
 
 	return MARSHAL_CLEAN;
 
+ demarshal_exception:
+	/* FIXME: may well leak */
+	CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
+				   CORBA_COMPLETED_MAYBE);
+	return MARSHAL_EXCEPTION_COMPLETE;
+
  msg_exception:
 	if (giop_recv_buffer_reply_status (recv_buffer) ==
 	    GIOP_LOCATION_FORWARD) {
@@ -582,7 +592,7 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 {
 	CORBA_unsigned_long     request_id;
 	CORBA_completion_status completion_status;
-	GIOPConnection         *cnx;
+	GIOPConnection         *cnx = NULL;
 	GIOPMessageQueueEntry   mqe;
 	ORBit_OAObject          adaptor_obj;
 
@@ -657,6 +667,8 @@ ORBit_small_invoke_stub (CORBA_Object       obj,
 
  clean_out:
 	tprintf_end_method ();
+	if (cnx)
+		linc_object_unref (cnx);
 	return;
 
  system_exception:
@@ -698,7 +710,7 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 		switch (tc->kind) {
 		case BASE_TYPES:
 		case OBJ_STRING_TYPES:
-			retval = alloca (ORBit_gather_alloc_info (tc));
+			retval = g_alloca (ORBit_gather_alloc_info (tc));
 			break;
 		case STRUCT_UNION_TYPES:
 			if (m_data->flags & ORBit_I_COMMON_FIXED_SIZE) {
@@ -716,8 +728,10 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 		int len = m_data->arguments._length *
 			sizeof (gpointer);
 
-		args = alloca (len);
-		scratch = alloca (len);
+		args = g_alloca (len);
+		memset (args, 0, len);
+		scratch = g_alloca (len);
+		memset (scratch, 0, len);
 	}		
 
 	for (i = 0; i < m_data->arguments._length; i++) {
@@ -735,7 +749,7 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 			switch (tc->kind) {
 			case BASE_TYPES:
 			case OBJ_STRING_TYPES:
-				p = args [i] = alloca (ORBit_gather_alloc_info (tc));
+				p = args [i] = g_alloca (ORBit_gather_alloc_info (tc));
 				do_demarshal_value (recv_buffer, &p, tc, orb);
 
 				p = args [i];
@@ -744,7 +758,7 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 			case STRUCT_UNION_TYPES:
 			case CORBA_tk_array:
 				if (a->flags & ORBit_I_COMMON_FIXED_SIZE) {
-					p = args [i] = alloca (ORBit_gather_alloc_info (tc));
+					p = args [i] = g_alloca (ORBit_gather_alloc_info (tc));
 					do_demarshal_value (recv_buffer, &p, tc, orb);
 					p = args [i];
 					tprintf_trace_value (&p, tc);
@@ -768,7 +782,7 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 			switch (tc->kind) {
 			case BASE_TYPES:
 			case OBJ_STRING_TYPES:
-				scratch [i] = alloca (ORBit_gather_alloc_info (tc));
+				scratch [i] = g_alloca (ORBit_gather_alloc_info (tc));
 				break;
 			case STRUCT_UNION_TYPES:
 			case CORBA_tk_array:
@@ -802,8 +816,14 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 
 	if (m_data->flags & ORBit_I_METHOD_1_WAY)
 		goto clean_out;
+	else
+		goto handle_possible_exception;
+	
+ demarshal_exception:
+	CORBA_exception_set_system (ev, ex_CORBA_MARSHAL,
+				    CORBA_COMPLETED_NO);
 
- sys_exception:
+ handle_possible_exception:
 	/* FIXME: should we be using the connection's GIOP version ? */
 	send_buffer = giop_send_buffer_use_reply (
 		recv_buffer->giop_version,
@@ -821,7 +841,7 @@ ORBit_small_invoke_adaptor (ORBit_OAObject     adaptor_obj,
 			   so we throw a system exception next */
 			dprintf (MESSAGES, "Re-sending an exception, this time %d: '%s'",
 				 ev->_major, ev->_id);
-			goto sys_exception;
+			goto handle_possible_exception;
 		}
 		tprintf ("User exception '%s'", ev->_id);
 
@@ -1076,7 +1096,7 @@ async_recv_cb (ORBitAsyncQueueEntry *aqe)
 	CORBA_exception_init (ev);
 
 	/* So we don't get invoked again */
-	aqe->mqe.u.unthreaded.cb = NULL;
+	aqe->mqe.async_cb = NULL;
 
 	if (!aqe->mqe.cnx ||
 	    aqe->mqe.cnx->parent.status == LINC_DISCONNECTED)
@@ -1157,6 +1177,8 @@ ORBit_small_invoke_async (CORBA_Object         obj,
 	aqe->m_data = /* ORBit_RootObject_duplicate */ (m_data);
 
  clean_out:
+	if (cnx)
+		linc_object_unref (cnx);
 	tprintf_end_method ();
 	return;
 
@@ -1187,13 +1209,13 @@ ORBit_small_get_servant (CORBA_Object obj)
 }
 
 static ORBitConnectionStatus
-get_status (LINCConnection *cnx)
+get_status (GIOPConnection *cnx)
 {
 	ORBitConnectionStatus ret;
 
 	g_return_val_if_fail (cnx != NULL, ORBIT_CONNECTION_DISCONNECTED);
 
-	switch (cnx->status) {
+	switch (linc_connection_get_status (LINC_CONNECTION (cnx))) {
 	case LINC_CONNECTED:
 		ret = ORBIT_CONNECTION_CONNECTED;
 		break;
@@ -1215,13 +1237,14 @@ ORBit_small_get_connection_status (CORBA_Object obj)
 	if (ORBit_small_get_servant (obj))
 		ret = ORBIT_CONNECTION_IN_PROC;
 	else {
-		GIOPConnection *connection;
+		GIOPConnection *cnx;
 
-		connection = ORBit_object_get_connection (obj);
+		cnx = ORBit_object_get_connection (obj);
 
-		if (connection)
-			ret = get_status (LINC_CONNECTION (connection));
-		else
+		if (cnx) {
+			ret = get_status (cnx);
+			linc_object_unref (cnx);
+		} else
 			ret = ORBIT_CONNECTION_DISCONNECTED;
 	}
 
@@ -1242,14 +1265,14 @@ ORBit_small_listen_for_broken (CORBA_Object obj,
 		ret = ORBIT_CONNECTION_IN_PROC;
 
 	else {
-		GIOPConnection *connection;
+		GIOPConnection *cnx;
 
-		connection = ORBit_object_get_connection (obj);
+		cnx = ORBit_object_get_connection (obj);
 
-		if (connection) {
-			ret = get_status (LINC_CONNECTION (connection));
-			g_signal_connect (G_OBJECT (connection),
-					  "broken", fn, user_data);
+		if (cnx) {
+			ret = get_status (cnx);
+			g_signal_connect (cnx, "broken", fn, user_data);
+			linc_object_unref (cnx);
 		} else
 			ret = ORBIT_CONNECTION_DISCONNECTED;
 	}
@@ -1270,16 +1293,16 @@ ORBit_small_unlisten_for_broken (CORBA_Object obj,
 		ret = ORBIT_CONNECTION_IN_PROC;
 
 	else {
-		GIOPConnection *connection;
+		GIOPConnection *cnx;
 
-		connection = ORBit_object_get_connection (obj);
+		cnx = ORBit_object_get_connection (obj);
 
-		if (connection) {
-			ret = get_status (LINC_CONNECTION (connection));
+		if (cnx) {
+			ret = get_status (cnx);
 			g_signal_handlers_disconnect_matched (
-				G_OBJECT (connection), 
-				G_SIGNAL_MATCH_FUNC,
+				cnx, G_SIGNAL_MATCH_FUNC,
 				0, 0, NULL, G_CALLBACK (fn), NULL);
+			linc_object_unref (cnx);
 		} else
 			ret = ORBIT_CONNECTION_DISCONNECTED;
 	}
@@ -1288,11 +1311,31 @@ ORBit_small_unlisten_for_broken (CORBA_Object obj,
 }
 
 ORBitConnection *
-ORBit_small_get_connection (CORBA_Object obj)
+ORBit_small_get_connection_ref (CORBA_Object obj)
 {
 	return (ORBitConnection *) ORBit_object_get_connection (obj);
 }
 
+ORBitConnection *
+ORBit_small_get_connection (CORBA_Object obj)
+{
+	ORBitConnection *cnx;
+
+	cnx = ORBit_small_get_connection_ref (obj);
+
+	/* This sucks but compatibily */
+	ORBit_small_connection_unref (cnx);
+	
+	return cnx;
+}
+
+void
+ORBit_small_connection_unref (ORBitConnection *cnx)
+{
+	if (cnx)
+		linc_object_unref (cnx);
+}
+
 void
 ORBit_connection_set_max_buffer (ORBitConnection *cnx,
 				 gulong           max_buffer_bytes)
