@@ -27,8 +27,10 @@
 #include "constants.h"
 #include "orb-core/orb-core-private.h"
 #include "orbit-imodule.h"
+#include "orbit/orbit.h"
 
 #define NUM_RUNS 1
+#define NUM_THREADS 0
 
 #undef TIMING_RUN
 
@@ -88,19 +90,13 @@ testSequenceHelpers (void)
 }
 
 static void
-testAttribute (test_TestFactory factory, 
+testAttribute (test_BasicServer   objref,
 	       CORBA_Environment *ev)
-{  
-	test_BasicServer objref;
-	CORBA_char      *val;
-	CORBA_long       lval;
+{  	
+	CORBA_char *val;
+	CORBA_long  lval;
 
 	d_print ("Testing attributes...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
-	g_assert (objref != CORBA_OBJECT_NIL);
-	g_assert (CORBA_Object_is_a (objref, "IDL:orbit/test/BasicServer:1.0", ev));
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	val = test_BasicServer__get_foo (objref, ev);
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);  
@@ -113,32 +109,26 @@ testAttribute (test_TestFactory factory,
 	lval = test_BasicServer__get_bah (objref, ev);
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);  
 	g_assert (lval == constants_LONG_RETN);
-
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testString (test_TestFactory   factory, 
+testString (test_BasicServer   objref,
 	    CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	const CORBA_char *in;
 	CORBA_char *inout, *out, *retn; 
 
 	d_print ("Testing strings...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	in = constants_STRING_IN;
 	inout = CORBA_string_dup (constants_STRING_INOUT_IN);
 	retn = test_BasicServer_opString (objref, in, &inout, &out, ev);
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
   
-	g_assert (strcmp (in, constants_STRING_IN)==0);
-	g_assert (strcmp (inout, constants_STRING_INOUT_OUT)==0);
 	g_assert (strcmp (out, constants_STRING_OUT)==0);
 	g_assert (strcmp (retn, constants_STRING_RETN)==0);	
+	g_assert (strcmp (in, constants_STRING_IN)==0);
+	g_assert (strcmp (inout, constants_STRING_INOUT_OUT)==0);
 
 	test_BasicServer_opOneWay (objref, constants_STRING_IN, ev);
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
@@ -146,21 +136,15 @@ testString (test_TestFactory   factory,
 	CORBA_free (inout);
 	CORBA_free (out);
 	CORBA_free (retn);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
-
 static void
-testLong (test_TestFactory   factory, 
+testLong (test_BasicServer   objref,
 	  CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	CORBA_long inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing longs...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	inArg = constants_LONG_IN;
 	inoutArg = constants_LONG_INOUT_IN;
@@ -170,21 +154,16 @@ testLong (test_TestFactory   factory,
 	g_assert (inoutArg == constants_LONG_INOUT_OUT);
 	g_assert (outArg == constants_LONG_OUT);
 	g_assert (retn == constants_LONG_RETN);
-
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testLongLong (test_TestFactory   factory, 
+testLongLong (test_BasicServer   objref,
 	      CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	CORBA_long_long inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing long longs...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
 	inArg = constants_LONG_LONG_IN;
 	inoutArg = constants_LONG_LONG_INOUT_IN;
 	retn = test_BasicServer_opLongLong (objref, inArg, &inoutArg, &outArg, ev);
@@ -193,20 +172,16 @@ testLongLong (test_TestFactory   factory,
 	g_assert (inoutArg == constants_LONG_LONG_INOUT_OUT);
 	g_assert (outArg == constants_LONG_LONG_OUT);
 	g_assert (retn == constants_LONG_LONG_RETN);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testFloat (test_TestFactory   factory, 
+testFloat (test_BasicServer   objref,
 	   CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	CORBA_float inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing floats...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
 	inArg = constants_FLOAT_IN;
 	inoutArg = constants_FLOAT_INOUT_IN;
 	retn = test_BasicServer_opFloat (objref, inArg, &inoutArg, &outArg, ev);
@@ -215,20 +190,16 @@ testFloat (test_TestFactory   factory,
 	g_assert (inoutArg == constants_FLOAT_INOUT_OUT);
 	g_assert (outArg == constants_FLOAT_OUT);
 	g_assert (retn == constants_FLOAT_RETN);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testDouble (test_TestFactory   factory, 
+testDouble (test_BasicServer   objref,
 	    CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	CORBA_double inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing doubles...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
 	inArg = constants_DOUBLE_IN;
 	inoutArg = constants_DOUBLE_INOUT_IN;
 	retn = test_BasicServer_opDouble (objref, inArg, &inoutArg, &outArg, ev);
@@ -237,20 +208,16 @@ testDouble (test_TestFactory   factory,
 	g_assert (inoutArg == constants_DOUBLE_INOUT_OUT);
 	g_assert (outArg == constants_DOUBLE_OUT);
 	g_assert (retn == constants_DOUBLE_RETN);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testLongDouble (test_TestFactory   factory, 
-	    CORBA_Environment *ev)
+testLongDouble (test_BasicServer   objref,
+		CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	CORBA_long_double inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing long doubles...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
 	inArg = constants_LONG_DOUBLE_IN;
 	inoutArg = constants_LONG_DOUBLE_INOUT_IN;
 	retn = test_BasicServer_opLongDouble (objref, inArg, &inoutArg, &outArg, ev);
@@ -259,20 +226,16 @@ testLongDouble (test_TestFactory   factory,
 	g_assert (inoutArg == constants_LONG_DOUBLE_INOUT_OUT);
 	g_assert (outArg == constants_LONG_DOUBLE_OUT);
 	g_assert (retn == constants_LONG_DOUBLE_RETN);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testEnum (test_TestFactory   factory, 
+testEnum (test_BasicServer   objref,
 	  CORBA_Environment *ev)
 {
-	test_BasicServer objref;
 	test_AnEnum inArg, inoutArg, outArg, retn;
 
 	d_print ("Testing enums...\n");
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
 	inArg = test_ENUM_IN;
 	inoutArg = test_ENUM_INOUT_IN;
 	retn = test_BasicServer_opEnum (objref, inArg, &inoutArg, &outArg, ev);
@@ -281,23 +244,17 @@ testEnum (test_TestFactory   factory,
 	g_assert (inoutArg == test_ENUM_INOUT_OUT);
 	g_assert (outArg == test_ENUM_OUT);
 	g_assert (retn == test_ENUM_RETN);
-	CORBA_Object_release (objref, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
 
 static void
-testException (test_TestFactory   factory, 
+testException (test_BasicServer   objref,
 	       CORBA_Environment *ev)
 {
-	test_BasicServer    objref;
 	test_TestException *ex;
-	CORBA_Environment  *cpyev, *rev = ev;
+	CORBA_Environment  *cpyev;
 
 	d_print ("Testing exceptions...\n");
 
-	objref = test_TestFactory_getBasicServer (factory, ev);
-	g_assert (ev->_major == CORBA_NO_EXCEPTION);
-
 	test_BasicServer_opException (CORBA_OBJECT_NIL, ev);
 	g_assert (ev->_major == CORBA_SYSTEM_EXCEPTION);
 	g_assert (strcmp (CORBA_exception_id (ev), ex_CORBA_INV_OBJREF) == 0);
@@ -329,9 +286,6 @@ testException (test_TestFactory   factory,
 	g_assert (ex->aseq._buffer[0] == constants_LONG_IN);*/
 
 	CORBA_free (cpyev);
-
-	CORBA_Object_release (objref, rev);
-	g_assert (rev->_major == CORBA_NO_EXCEPTION);
 }
 
 static CORBA_TypeCode
@@ -585,7 +539,7 @@ testUnboundedSequence (test_TestFactory   factory,
 		       CORBA_Environment *ev)
 {
   test_SequenceServer objref;
-  test_StrSeq inArg, inoutArg, *outArg, *retn;
+  test_StrSeq *outArg = NULL, inArg, inoutArg, *retn;
   guint i;
   d_print ("Testing unbounded sequences...\n");
   objref = test_TestFactory_getSequenceServer (factory, ev);
@@ -815,17 +769,12 @@ testMiscUnions (test_TestFactory   factory,
 }
 
 static void
-testFixedLengthArray (test_TestFactory   factory, 
-		      CORBA_Environment *ev)
+testLongArray (test_ArrayServer   objref,
+	       CORBA_Environment *ev)
 {
-  
-  test_ArrayServer objref;
+  int i;
   test_LongArray inArg, inoutArg, outArg;
   test_LongArray_slice *retn;
-  int i;
-  d_print ("Testing arrays with fixed length members...\n");
-  objref = test_TestFactory_getArrayServer (factory, ev);
-  g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
   for (i=0;i<test_SequenceLen;i++)
 	inArg[i] = constants_SEQ_LONG_IN[i];
@@ -846,6 +795,50 @@ testFixedLengthArray (test_TestFactory   factory,
 	g_assert (retn[i]==constants_SEQ_LONG_RETN[i]);
 
   CORBA_free (retn);
+}
+
+static void
+testOctetArray (test_ArrayServer   objref,
+		CORBA_Environment *ev)
+{
+  int i;
+  test_OctetArray inArg, inoutArg, outArg;
+  test_OctetArray_slice *retn;
+
+  for (i=0;i<test_SequenceLen;i++)
+	inArg[i] = constants_SEQ_OCTET_IN[i];
+
+  for (i=0;i<test_SequenceLen;i++)
+	inoutArg[i] = constants_SEQ_OCTET_INOUT_IN[i];
+
+  retn = test_ArrayServer_opOctetArray (objref, inArg, inoutArg, outArg, ev);
+  g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+  for (i=0;i<test_SequenceLen;i++)
+	g_assert (inArg[i]==constants_SEQ_OCTET_IN[i]);
+  for (i=0;i<test_SequenceLen;i++)
+	g_assert (inoutArg[i]==constants_SEQ_OCTET_INOUT_OUT[i]);
+  for (i=0;i<test_SequenceLen;i++)
+	g_assert (outArg[i]==constants_SEQ_OCTET_OUT[i]);
+  for (i=0;i<test_SequenceLen;i++)
+	g_assert (retn[i]==constants_SEQ_OCTET_RETN[i]);
+
+  CORBA_free (retn);
+}
+
+static void
+testFixedLengthArray (test_TestFactory   factory, 
+		      CORBA_Environment *ev)
+{
+  
+  test_ArrayServer objref;
+  d_print ("Testing arrays with fixed length members...\n");
+  objref = test_TestFactory_getArrayServer (factory, ev);
+  g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+  testLongArray (objref, ev);
+  testOctetArray (objref, ev);
+
   CORBA_Object_release (objref, ev);
   g_assert (ev->_major == CORBA_NO_EXCEPTION);
 }
@@ -1566,7 +1559,7 @@ testAsync (test_TestFactory   factory,
 
 	if (in_proc) {
 		g_assert (objref->profile_list == NULL);
-		g_assert (ORBit_object_get_connection (objref) == NULL);
+		g_assert (ORBit_small_get_connection_ref (objref) == NULL);
 	}
 			      
 	done = 0;
@@ -1576,7 +1569,7 @@ testAsync (test_TestFactory   factory,
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	/* While waiting do some normal methods */
-	testString (factory, ev);
+	testString (objref, ev);
 
 	wait_until_done ();
 
@@ -1667,7 +1660,7 @@ testPingPong (test_TestFactory   factory,
 
 	if (!in_proc) {
 		int i;
-		ORBitConnection *cnx = ORBit_small_get_connection (r_objref);
+		ORBitConnection *cnx = ORBit_small_get_connection_ref (r_objref);
 		const char *base =
 			"This string is in order to provide some "
 			"more bulky data on the wire, the larger "
@@ -1708,8 +1701,9 @@ testPingPong (test_TestFactory   factory,
 
 		/* If this blows - perhaps you just have a strange
 		 * system scheduler */
-//		g_assert (broken);
+/*		g_assert (broken); */
 		CORBA_exception_free (ev);
+		ORBit_small_connection_unref (cnx);
 	}
 
 	CORBA_Object_release (objref, ev);
@@ -1836,8 +1830,35 @@ test_time_noop (test_TestFactory   factory,
 	ORBit_small_flags = old_flags;
 }
 
+static void
+test_basic_server (test_TestFactory   factory, 
+		   CORBA_Environment *ev)
+{
+	test_BasicServer objref;
+
+	objref = test_TestFactory_getBasicServer (factory, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (objref != CORBA_OBJECT_NIL);
+	g_assert (CORBA_Object_is_a (objref, "IDL:orbit/test/BasicServer:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	testAttribute  (objref, ev);
+	testString     (objref, ev);
+	testLong       (objref, ev);
+	testLongLong   (objref, ev);
+	testFloat      (objref, ev);
+	testDouble     (objref, ev);
+	testLongDouble (objref, ev);
+	testEnum       (objref, ev);
+	testException  (objref, ev);
+
+	CORBA_Object_release (objref, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+}
+
 static void
 run_tests (test_TestFactory   factory, 
+	   gboolean           reenterant_only,
 	   CORBA_Environment *ev)
 {
 	int i;
@@ -1845,15 +1866,7 @@ run_tests (test_TestFactory   factory,
 	for (i = 0; i < NUM_RUNS; i++) {
 		testSequenceHelpers ();
 		testConst ();
-		testAttribute (factory, ev);
-		testString (factory, ev);
-		testLong (factory, ev);
-		testLongLong (factory, ev);
-		testFloat (factory, ev);
-		testDouble (factory, ev);
-		testLongDouble (factory, ev);
-		testEnum (factory, ev);
-		testException (factory, ev);
+		test_basic_server (factory, ev);
 		testIsA (factory, ev);
 		testFixedLengthStruct (factory, ev);
 		testVariableLengthStruct (factory, ev);
@@ -1877,15 +1890,13 @@ run_tests (test_TestFactory   factory,
 		testTypeCode (factory, ev);
 		testContext (factory, ev);
 		testIInterface (factory, ev);
-		testMisc (factory, ev);
 #ifndef TIMING_RUN
-		testAsync (factory, ev);
+		if (!reenterant_only)
+			testAsync (factory, ev);
 #endif
 		if (!in_proc)
 			testPingPong (factory, ev);
-#if NUM_RUNS == 1
-		testSegv (factory, ev);
-#endif
+		testMisc (factory, ev);
 	}
 	
 #if NUM_RUNS > 1
@@ -1893,6 +1904,46 @@ run_tests (test_TestFactory   factory,
 #endif
 }
 
+static gpointer
+test_thread (gpointer data)
+{
+	CORBA_Environment ev[1];
+	test_TestFactory factory = data;
+
+	CORBA_exception_init (ev);
+	run_tests (factory, TRUE, ev);
+	CORBA_exception_free (ev);
+
+	return data;
+}
+
+static void
+run_threaded_tests (test_TestFactory   factory, 
+		    CORBA_Environment *ev)
+{
+	int i;
+	GError *error = NULL;
+	GThread **threads;
+
+	if (!NUM_THREADS)
+		return;
+
+	fprintf (stderr, "Testing with %d threads\n", NUM_THREADS);
+
+	threads = g_new0 (GThread *, NUM_THREADS);
+
+	for (i = 0; i < NUM_THREADS; i++) {
+		threads [i] = g_thread_create
+			( test_thread, factory, TRUE, &error);
+		g_assert (!error);
+	}
+
+	for (i = 0; i < NUM_THREADS; i++) {
+		if (!(g_thread_join (threads [i]) == factory))
+			g_error ("Wierd thread join problem '%d'", i);
+	}
+}
+
 static void
 dump_protos (void)
 {
@@ -1927,7 +1978,7 @@ test_init (CORBA_Environment *ev)
 int
 main (int argc, char *argv [])
 {
-	CORBA_Environment  ev;
+	CORBA_Environment  ev[1];
 	test_TestFactory   factory;
 	ORBit_IInterfaces *interfaces = NULL;
 	gboolean           gen_imodule = FALSE;
@@ -1935,13 +1986,13 @@ main (int argc, char *argv [])
 	const char        *orb_name;
 	int                i;
 
-	CORBA_exception_init (&ev);
+	CORBA_exception_init (ev);
 
 /* FIXME - make this work nicely sometime.
-	global_orb = CORBA_ORB_init (&argc, argv, "", &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
-	CORBA_Object_release (global_orb, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	global_orb = CORBA_ORB_init (&argc, argv, "", ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	CORBA_Object_release (global_orb, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 */
 	for (i = 0; i < argc; i++) {
 		if (!strcmp (argv [i], "--gen-imodule"))
@@ -1955,8 +2006,8 @@ main (int argc, char *argv [])
 	else
 		orb_name = "orbit-local-orb";
 
-	global_orb = CORBA_ORB_init (&argc, argv, orb_name, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	global_orb = CORBA_ORB_init (&argc, argv, orb_name, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	if (gen_imodule) {
 		CORBA_sequence_CORBA_TypeCode *typecodes = NULL;
@@ -1968,11 +2019,11 @@ main (int argc, char *argv [])
 
 		CORBA_free (typecodes);
 
-		init_iinterfaces (interfaces, &ev);
+		init_iinterfaces (interfaces, ev);
 	}
 
-	test_init (&ev);
-	test_initial_references (global_orb, &ev);
+	test_init (ev);
+	test_initial_references (global_orb, ev);
 
 	free (malloc (8)); /* -lefence */
 
@@ -1982,15 +2033,17 @@ main (int argc, char *argv [])
 	in_proc = TRUE;
 
 	fprintf (stderr, "\n --- In proc ---\n\n\n");
-	factory = get_server (global_orb, &ev);
+	factory = get_server (global_orb, ev);
 	g_assert (factory->profile_list == NULL);
 	g_assert (ORBit_object_get_connection (factory) == NULL);
 
-	test_time_noop (factory, &ev);
-	run_tests (factory, &ev);
+	test_time_noop (factory, ev);
+	run_tests (factory, FALSE, ev);
+	if (threaded)
+		run_threaded_tests (factory, ev);
 
-	CORBA_Object_release (factory, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	CORBA_Object_release (factory, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 	factory = CORBA_OBJECT_NIL;
 
 	fprintf (stderr, "\n\n --- Out of proc ---\n\n\n");
@@ -2008,31 +2061,34 @@ main (int argc, char *argv [])
 		fclose (infile);
 		ior [size] = '\0';   /* insure that string is terminated correctly */
 
-		factory = CORBA_ORB_string_to_object (global_orb, ior, &ev);
-		g_assert (ev._major == CORBA_NO_EXCEPTION);
+		factory = CORBA_ORB_string_to_object (global_orb, ior, ev);
+		g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
-		if (CORBA_Object_non_existent (factory, &ev))
+		if (CORBA_Object_non_existent (factory, ev))
 			g_error  ("Can't contact the server");
-		g_assert (ev._major == CORBA_NO_EXCEPTION);
+		g_assert (ev->_major == CORBA_NO_EXCEPTION);
 	}
-	run_tests (factory, &ev);
+	run_tests (factory, FALSE, ev);
+	if (threaded)
+		run_threaded_tests (factory, ev);
+	testSegv (factory, ev);
 
-	CORBA_Object_release (factory, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	CORBA_Object_release (factory, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
 	if (gen_imodule)
 		CORBA_free (interfaces);
 
-	CORBA_Object_release ((CORBA_Object) global_poa, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	CORBA_Object_release ((CORBA_Object) global_poa, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 	
-	CORBA_ORB_destroy (global_orb, &ev);
-	CORBA_exception_free (&ev);
+	CORBA_ORB_destroy (global_orb, ev);
+	CORBA_exception_free (ev);
 
-	CORBA_Object_release ((CORBA_Object) global_orb, &ev);
-	g_assert (ev._major == CORBA_NO_EXCEPTION);
+	CORBA_Object_release ((CORBA_Object) global_orb, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 
-	CORBA_exception_free (&ev);
+	CORBA_exception_free (ev);
 
 	d_print ("All tests passed successfully\n");
 
